/* Copyright (c) 1995 by Stanford University.  All Rights Reserved */

/***************************************************************************
   NAME
     mif_init.c
   PURPOSE
     Sample initialization code for CS326 projects
   NOTES
     
   AUTHOR
     Tsai-Yen Li (li@flamingo.Stanford.EDU)
   HISTORY
     Tsai-Yen Li - Apr 10, 1995: Created.
***************************************************************************/
#ifdef RCS
#ifndef lint
 static char rcsid[]="$Header: /u/li/Proj/mif/RCS/mif_init.c,v 1.4 1995/04/15 22:07:52 li Exp li $";
#endif 
#endif 
/* $Log: mif_init.c,v $
 * Revision 1.4  1995/04/15  22:07:52  li
 * add path interface
 *
 * Revision 1.3  1995/04/14  01:33:14  li
 * Deal with multiple robots
 *
 * Revision 1.2  1995/04/12  08:00:18  li
 * add comments and fix some bugs
 *
 * Revision 1.1  1995/04/11  21:15:40  li
 * Initial revision
 * */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/param.h>
#include <time.h>
#include <X11/keysym.h>
#include <Xm/Xm.h>
#include <Xm/Scale.h>
#include <Xm/MainW.h>
#include <Mrm/MrmPublic.h>
#include "planner_header.h"
#include "mif_header.h"
#include "mif_proto.h"
#include "parse.h"

static char ch;
#define fscanf_comment(fd)   {do {ch=fgetc(fd);} while(ch != '\n');}

/*****************************************************************************
  NO_MORE_MEMORY. Can't allocate more memory. Quit.
******************************************************************************/

void 
NO_MORE_MEMORY(char *calling_routine) {

  fprintf(logfd,"In %s\n",calling_routine);
  fprintf(logfd,"Can't allocate more memory space\n");
  exit(-1);

}

/*****************************************************************************
  Allocate_Obstacle: Allocate memory for obstacle data 
******************************************************************************/

obst_data *
Allocate_Obstacle(int nobst) {

  obst_data *ret;

  if (nobst==0) return(NULL);
  if ((ret=(obst_data *)calloc(nobst,sizeof(obst_data)))==NULL) {
    NO_MORE_MEMORY("Allocate_Obstacle");
    exit(-1);
  }
  return(ret);

}

/*****************************************************************************
  Allocate_Robot: Allocate memory for robot data 
******************************************************************************/

robot_data *
Allocate_Robot(int nrobot) {

  robot_data *ret;

  if (nrobot==0) return(NULL);
  if ((ret=(robot_data *)calloc(nrobot,sizeof(robot_data)))==NULL) {
    NO_MORE_MEMORY("Allocate_Robot");
    exit(-1);
  }
  return(ret);

}

/*****************************************************************************
  Allocate_Fpolygon: Allocating a given number of Fpolygon
******************************************************************************/

Fpolygon *
Allocate_Fpolygon(int npoly) {

  Fpolygon *ret;

  if (npoly==0) return(NULL);
  if ((ret=(Fpolygon *)calloc(npoly,sizeof(Fpolygon)))==NULL) {
    NO_MORE_MEMORY("Allocate_Fpolygon");
    exit(-1);
  }
  return(ret);
}

/*****************************************************************************
  Allocate_Ipolygon: Allocating a given number of Ipolygon
******************************************************************************/

Ipolygon *
Allocate_Ipolygon(int npoly) {

  Ipolygon *ret;

  if (npoly==0) return(NULL);
  if ((ret=(Ipolygon *)calloc(npoly,sizeof(Ipolygon)))==NULL) {
    NO_MORE_MEMORY("Allocate_Ipolygon");
    exit(-1);
  }
  return(ret);
}

/*****************************************************************************
  Allocate_Ivert: Allocating a given number of vertices for a Ipolygon
******************************************************************************/

Ipoint *
Allocate_Ivert(int nvert) {

  Ipoint *ret;

  if (nvert==0) return(NULL);
  if ((ret=(Ipoint *)calloc(nvert+1,sizeof(Ipoint)))==NULL) {
    NO_MORE_MEMORY("Allocate_Ivert");
    exit(-1);
  }
  return(ret);
}

/*****************************************************************************
  Allocate_Fvert: Allocating a given number of vertices for a Fpolygon
******************************************************************************/

Fpoint *
Allocate_Fvert(int nvert) {

  Fpoint *ret;

  if (nvert==0) return(NULL);
  if ((ret=(Fpoint *)calloc(nvert+1,sizeof(Fpoint)))==NULL) {
    NO_MORE_MEMORY("Allocate_Fvert");
    exit(-1);
  }
  return(ret);
}

/*****************************************************************************
  Read_Robots: Initialize the robot data from a description data
******************************************************************************/

void 
Read_Robots(char *filename,int *nrobot,robot_data **robots) {

  FILE *fd;
  register int i,j,k;
  float x,y;
  int minx=5000,maxx=0,miny=5000,maxy=0,nvert,npoly;
  planner_obj *pobj;
  canvas_obj *cobj;
  canvas_obj *cgobj;

  fd=fopen(filename,"r");
  
  if (fd == NULL) {
    fprintf(logfd,"Error opening file %s\n",filename);
    return;
  } 

  V1(fprintf(logfd,"Reading robot data from file %s\n",filename));
  /* initial configuration */
  fscanf_comment(fd);
  fscanf(fd,"%d\n",nrobot);
  V2(fprintf(logfd,"Number of robots:%d\n",*nrobot));
  
  (*robots)=Allocate_Robot(*nrobot);
  probot=(planner_obj **) calloc(*nrobot,sizeof(planner_obj *));

  for(k=0;k<*nrobot;k++) {

    V2(fprintf(logfd,"Robots #%d\n",k));
    fscanf_comment(fd);
    probot[k]= &((*robots)[k].pobj);
    pobj= &(*robots)[k].pobj;
    cobj= &(*robots)[k].cobj;
    cgobj= &(*robots)[k].cgobj;
    
    /* init configuration */
    V2(fprintf(logfd,"initial configuration:\n"));
    fscanf_comment(fd);
    fscanf(fd,"%f %f %f\n",&pobj->init.x,&pobj->init.y,&pobj->init.th);
    pobj->init.x *= wscale;
    pobj->init.y *= wscale;
    pobj->init.th=RNDINT(pobj->init.th/360.0*dth);
    V2(fprintf(logfd,"%f %f %f\n",pobj->init.x,pobj->init.y,pobj->init.th));
    
    /* goal configuration */
    V2(fprintf(logfd,"goal configuration:\n"));
    fscanf_comment(fd);
    fscanf(fd,"%f %f %f\n",&pobj->goal.x,&pobj->goal.y,&pobj->goal.th);
    pobj->goal.x *= wscale;
    pobj->goal.y *= wscale;
    pobj->goal.th=RNDINT(pobj->goal.th/360.0*dth);
    V2(fprintf(logfd,"%f %f %f\n",pobj->goal.x,pobj->goal.y,pobj->goal.th));
    
    /* geometric description */
    fscanf_comment(fd); /* number of vertices */
    fscanf(fd,"%d\n",&pobj->npoly);
    V2(fprintf(logfd,"Number of polygons: %d\n",pobj->npoly));
    npoly=cobj->npoly=cgobj->npoly=pobj->npoly;
    pobj->poly  = Allocate_Fpolygon(npoly);
    pobj->orig_poly = Allocate_Fpolygon(npoly);
    cobj->cpoly = Allocate_Ipolygon(npoly);
    cgobj->cpoly= Allocate_Ipolygon(npoly);
    cobj->ftpoly = Allocate_Fpolygon(npoly);
    cobj->itpoly = Allocate_Ipolygon(npoly);
    cobj->orig_poly_ptr = &pobj->orig_poly;
    
    for(i=0;i<pobj->npoly;i++) {
      
      fscanf_comment(fd); /* polygon number */
      V2(fprintf(logfd,"polygon #%d\n",i));
      fscanf_comment(fd); /* number of vertices */
      fscanf(fd,"%d\n",&pobj->orig_poly[i].nvert);
      V2(fprintf(logfd,"Number of vertices: %d\n",pobj->orig_poly[i].nvert));
      nvert=cobj->cpoly[i].nvert=cobj->itpoly[i].nvert=cobj->ftpoly[i].nvert=
	cgobj->cpoly[i].nvert=pobj->poly[i].nvert=pobj->orig_poly[i].nvert;
      /* allocating memory */
      pobj->orig_poly[i].vert = Allocate_Fvert(nvert);
      pobj->poly[i].vert = Allocate_Fvert(nvert);
      cobj->cpoly[i].vert = Allocate_Ivert(nvert);
      cobj->itpoly[i].vert = Allocate_Ivert(nvert);
      cobj->ftpoly[i].vert = Allocate_Fvert(nvert);
      cgobj->cpoly[i].vert = Allocate_Ivert(nvert);
      
      fscanf_comment(fd); /* number of vertices */
      for(j=0;j<pobj->orig_poly[i].nvert;j++) {
	fscanf(fd,"%f %f\n",&x,&y);
	x *= wscale;
	y *= wscale;
	minx=MIN(minx,x);
	maxx=MAX(maxx,x);
	miny=MIN(miny,y);
	maxy=MAX(maxy,y);
	pobj->orig_poly[i].vert[j].x=x;
	pobj->orig_poly[i].vert[j].y=y;
	
	V2(fprintf(logfd,"%f %f\n",pobj->orig_poly[i].vert[j].x,
		   pobj->orig_poly[i].vert[j].y));
      }
      pobj->orig_poly[i].vert[j]=pobj->orig_poly[i].vert[0];
    }
  }

  fclose(fd);

  return;
  
}


/*****************************************************************************
  Read_Obstacle: Initialize the obstacle data from a description data
******************************************************************************/

void 
Read_Obstacles(char *filename,int *nobst,obst_data **obst) {

  FILE *fd;
  register int i,j,k;
  float x,y;
  planner_obj *pobst;
  canvas_obj *cobst;

  fd=fopen(filename,"r");
  
  if (fd == NULL) {
    fprintf(logfd,"Error opening file %s\n",filename);
    return;
  } 

  V1(fprintf(logfd,"Reading Obstacle data from file %s\n",filename));
  fscanf_comment(fd);
  fscanf(fd,"%d\n",nobst);
  V2(fprintf(logfd,"Number of obstacles:%d\n",*nobst));
  (*obst)=Allocate_Obstacle(*nobst);
  pobstacle=(planner_obj **) calloc(*nobst,sizeof(planner_obj *));

  for(i=0;i<(*nobst);i++) {
    pobstacle[i]= &((*obst)[i].pobj);
    pobst= &((*obst)[i].pobj);
    cobst= &((*obst)[i].cobj);
    fscanf_comment(fd); /* obstacle number */
    V2(fprintf(logfd,"Obstacle #%d\n",i));
    /* initial configuration */
    fscanf_comment(fd); /* init configuration */
    fscanf(fd,"%f %f %f\n",&pobst->init.x,&pobst->init.y,&pobst->init.th);
    pobst->init.x *= wscale;
    pobst->init.y *= wscale;
    pobst->init.th=RNDINT(pobst->init.th/360.0*dth);
    V2(fprintf(logfd,"%f %f %f\n",pobst->init.x,pobst->init.y,
	    pobst->init.th));

    fscanf_comment(fd); /* number of polygons */
    fscanf(fd,"%d\n",&pobst->npoly);
    V2(fprintf(logfd,"Number of polygons:%d\n",pobst->npoly));
    cobst->npoly=pobst->npoly;
    pobst->poly = Allocate_Fpolygon(pobst->npoly);
    pobst->orig_poly = Allocate_Fpolygon(pobst->npoly);
    cobst->cpoly = Allocate_Ipolygon(cobst->npoly);
    cobst->orig_poly_ptr = &pobst->orig_poly; 
  
    for(j=0;j<pobst->npoly;j++) {
      fscanf_comment(fd); /* polygon number */
      fscanf_comment(fd); /* number of vertices */
      fscanf(fd,"%d\n",&pobst->orig_poly[j].nvert);
      V2(fprintf(logfd,"number of vertices: %d\n",pobst->orig_poly[j].nvert));
      cobst->cpoly[j].nvert=pobst->poly[j].nvert=pobst->orig_poly[j].nvert;
      /* allocate memory */
      pobst->poly[j].vert = Allocate_Fvert(pobst->poly[j].nvert);
      pobst->orig_poly[j].vert = Allocate_Fvert(pobst->poly[j].nvert);
      cobst->cpoly[j].vert = Allocate_Ivert(cobst->cpoly[j].nvert);

      fscanf_comment(fd); /* vertices */
      for(k=0;k<pobst->poly[j].nvert;k++) {
	fscanf(fd,"%f %f\n",&x,&y);
	x *= wscale;
	y *= wscale;
	pobst->orig_poly[j].vert[k].x=x;
	pobst->orig_poly[j].vert[k].y=y;
	V2(fprintf(logfd,"%f %f\n",pobst->orig_poly[j].vert[k].x,
		   pobst->orig_poly[j].vert[k].y));
      }
      pobst->orig_poly[j].vert[k]=pobst->orig_poly[j].vert[0];
    }
  }
  fclose(fd);

  return;
}

/*****************************************************************************
  Write_Robots: Write robot data into a description file
******************************************************************************/

void 
Write_Robots(char *filename,int nrobots,robot_data *robot) {

  FILE *fd;
  register int i,j,k;
  planner_obj *pobj;
  canvas_obj *cobj;
  canvas_obj *cgobj;

  fd=fopen(filename,"w");
  
  if (fd == NULL) {
    fprintf(logfd,"Error opening file %s\n",filename);
    return;
  } 

  V1(fprintf(logfd,"Writing the current robot data to file:%s\n",filename));

  /* initial configuration */
  fprintf(fd,"# number of robots\n");
  fprintf(fd,"%d\n",nrobots);
  
  for(k=0;k<nrobots;k++) {

    fprintf(fd,"# robot #%d\n",k);
    pobj= &robots[k].pobj;
    cobj= &robots[k].cobj;
    cgobj= &robots[k].cgobj;

    fprintf(fd,"# initial configuration\n");
    fprintf(fd,"%f %f %f\n",pobj->init.x/wscale,pobj->init.y/wscale,
	    pobj->init.th*360.0/(float)dth);
    
    fprintf(fd,"# goal configuration\n");
    fprintf(fd,"%f %f %f\n",pobj->goal.x/wscale,pobj->goal.y/wscale,
	    pobj->goal.th*360.0/(float)dth);
    
    /* geometric description */
    fprintf(fd,"# number of polygons\n");
    fprintf(fd,"%d\n",pobj->npoly);

    for(i=0;i<pobj->npoly;i++) {
      fprintf(fd,"# polygon #%d\n",i); /* polygon number */
      fprintf(fd,"# number of vertices\n"); /* number of vertices */
      fprintf(fd,"%d\n",pobj->orig_poly[i].nvert);
      fprintf(fd,"# vertices\n"); /* vertices */
      for(j=0;j<pobj->orig_poly[i].nvert;j++) {
	fprintf(fd,"%f %f\n",pobj->orig_poly[i].vert[j].x/wscale,
		pobj->orig_poly[i].vert[j].y/wscale);
      }
    }
  }

  fclose(fd);
  return;

}

/*****************************************************************************
  Write_Obstacles: Write obstacle data into a description file
******************************************************************************/

void 
Write_Obstacles(char *filename,int nobsts,obst_data *obst) {

  FILE *fd;
  register int i,j,k;
  planner_obj *pobst;
  canvas_obj *cobst;

  fd=fopen(filename,"w");
  
  if (fd == NULL) {
    fprintf(logfd,"Error opening file %s\n",filename);
    return;
  } 

  V1(fprintf(logfd,"Writing the current obstacle data to file:%s\n",filename));
  fprintf(fd,"number of obstacles\n");
  fprintf(fd,"%d\n",nobsts);

  for(i=0;i<nobsts;i++) {
    pobst= &obst[i].pobj;
    cobst= &obst[i].cobj;
    fprintf(fd,"# obstacle #%d\n",i); /* obstacle number */
    /* initial configuration */
    fprintf(fd,"# initial configuration\n"); /* init configuration */
    fprintf(fd,"%f %f %f\n",pobst->init.x/wscale,pobst->init.y/wscale,
	    pobst->init.th*360.0/(float)dth);
    fprintf(fd,"# number of polygons\n");
    fprintf(fd,"%d\n",pobst->npoly);

    for(j=0;j<pobst->npoly;j++) {
      fprintf(fd,"# polygon #%d\n",j); /* polygon number */
      fprintf(fd,"# number of vertices\n"); /* number of vertices */
      fprintf(fd,"%d\n",pobst->orig_poly[j].nvert);
      fprintf(fd,"# vertices\n"); /* vertices */
      for(k=0;k<pobst->poly[j].nvert;k++) {
	fprintf(fd,"%f %f\n",pobst->orig_poly[j].vert[k].x/wscale,
		pobst->orig_poly[j].vert[k].y/wscale);
      }
    }
  }
  fclose(fd);
  
  return;
}

/****************************************************************************
  Read_Paths: Write all paths to a file
****************************************************************************/

void 
Read_Paths(char *name,int nrobots,robot_data *robots) {

  register int i,j;
  objpath *curpath;
  int npath;
  float x,y,th;
  FILE *fd;
  
  fd=fopen(name,"r");
  if (fd == NULL) {
    fprintf(logfd,"Error opening file %s\n",name);
    return;
  }
  
  holo_path_exist=FALSE;
  nholo_path_exist=FALSE;
  fscanf_comment(fd); /* number of robots */
  fscanf(fd,"%d\n",&npath);
  if (npath!=nrobots) {
    fprintf(logfd,"The number of paths does not match the number of robots\n");
    fprintf(logfd,"Abort ... \n");
    fclose(fd);
    return;
  }
  fscanf_comment(fd); /* holonomic paths */
  for(i=0;i<nrobots;i++) {
    curpath= &robots[i].pobj.holo_path;
    fscanf_comment(fd); /* robot number */
    fscanf_comment(fd); /* path length */
    fscanf(fd,"%d\n",&curpath->len);
    if (curpath->len>0) {
      holo_path_exist=TRUE;
      Check_Path_Memory(curpath,curpath->len);
      for (j=0;j<curpath->len;j++) {
	fscanf(fd,"%f %f %f\n",&x,&y,&th);
	curpath->path[j].x=x*wscale,
	curpath->path[j].y=y*wscale,
	curpath->path[j].th=th/180.0*M_PI;
      }
    }
  }
  fscanf_comment(fd); /* nonholonomic paths */
  for(i=0;i<nrobots;i++) {
    curpath= &robots[i].pobj.nholo_path;
    fscanf_comment(fd); /* robot number */
    fscanf_comment(fd); /* path length */
    fscanf(fd,"%d\n",&curpath->len);
    if (curpath->len>0) {
      nholo_path_exist=TRUE;
      Check_Path_Memory(curpath,curpath->len);
      for (j=0;j<curpath->len;j++) {
	fscanf(fd,"%f %f %f\n",&x,&y,&th);
	curpath->path[j].x=x*wscale,
	curpath->path[j].y=y*wscale,
	curpath->path[j].th=th/180.0*M_PI;
      }
    }
  }
  fclose(fd);

}

/****************************************************************************
  Write_Paths: Write all paths to a file
****************************************************************************/

void 
Write_Paths(char *name,int nrobots,robot_data *robots) {

  register int i,j;
  objpath *curpath;

  FILE *fd;
  
  fd=fopen(name,"w");
  if (fd == NULL) {
    fprintf(logfd,"Error opening file %s\n",name);
    return;
  }
  
  fprintf(fd,"# number of paths\n"); /* number of robots */
  fprintf(fd,"%d\n",nrobots);
  fprintf(fd,"# holonomic paths\n"); 
  for(i=0;i<nrobots;i++) {
    curpath= &robots[i].pobj.holo_path;
    fprintf(fd,"# robot #%d\n",i);
    fprintf(fd,"# path length\n");
    fprintf(fd,"%d\n",curpath->len);
    for (j=0;j<curpath->len;j++) {
      fprintf(fd,"%f %f %f\n",curpath->path[j].x/wscale,
	      curpath->path[j].y/wscale,
	      curpath->path[j].th*180.0/M_PI);
    }
  }
  fprintf(fd,"# nonholonomic paths\n"); 
  for(i=0;i<nrobots;i++) {
    curpath= &robots[i].pobj.nholo_path;
    fprintf(fd,"# robot #%d\n",i);
    fprintf(fd,"# path length\n");
    fprintf(fd,"%d\n",curpath->len);
    for (j=0;j<curpath->len;j++) {
      fprintf(fd,"%f %f %f\n",curpath->path[j].x/wscale,
	      curpath->path[j].y/wscale,
	      curpath->path[j].th*180.0/M_PI);
    }
  }
  fclose(fd);
}

/*****************************************************************************
  Init_Planner: Initialize misc stuff here
*****************************************************************************/

void 
Init_Planner(int dx,int dy,int dth) {

  dth2rad=M_PI/(dth*0.5);
  rad2dth=(dth*0.5)/M_PI;
  delta_th=TWOPI/(float)dth;

  /* put other misc initialization stuff here */

}
