/* Copyright (c) 1995 by Stanford University.  All Rights Reserved */

/***************************************************************************
   NAME
     mif_main.c
   PURPOSE
     sample interface for CS326
   NOTES
     
   AUTHOR
     Tsai-Yen Li (li@flamingo.Stanford.EDU)
   HISTORY
     Tsai-Yen Li - Apr 9, 1995: Created.
***************************************************************************/
#ifdef RCS
#ifndef lint
 static char rcsid[]="$Header: /u/li/Proj/mif/RCS/mif_main.c,v 1.4 1995/04/15 22:07:52 li Exp li $";
#endif 
#endif 

/* $Log: mif_main.c,v $
 * Revision 1.4  1995/04/15  22:07:52  li
 * add path interface
 *
 * Revision 1.3  1995/04/14  01:33:14  li
 * Deal with multiple robots
 *
 * Revision 1.2  1995/04/12  08:00:18  li
 * add comments and fix some bugs
 *
 * Revision 1.1  1995/04/11  21:15:40  li
 * Initial revision
 * */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/param.h>
#include <time.h>
#include <X11/keysym.h>
#include <Xm/Xm.h>
#include <Xm/Scale.h>
#include <Xm/MainW.h>
#include <Mrm/MrmPublic.h>
#include "planner_header.h"
#include "mif_header.h"
#include "parse.h"
#include "mif_proto.h"

/****************************************************************************
  Global Variables
****************************************************************************/

/* general purpose globals */
char obst_filename[MAX_STR],robot_filename[MAX_STR],path_filename[MAX_STR];
int holo_path_exist=FALSE,nholo_path_exist=FALSE;
int verbosity=1,confirm=TRUE;
int debug_smooth=FALSE;
float wscale=1.0;
float pc_scale,cp_scale,rad2dth,dth2rad,delta_th;

/* Xt or Xm stuff */
timeout_data t_data; /* data struct for animation */
drawing_data canvas_data; /* for canvas attributes */
XtAppContext app_context; /* main pointer to the interface */
Widget shell,canvas; /* widgets called by users */
Widget loadRobotFileDialog,loadObstFileDialog,loadPathFileDialog;
Widget saveRobotFileDialog,saveObstFileDialog,savePathFileDialog;
Widget infoDialog,confirmDialog,commandL,planL;

/* planner globals */
int nrobots;
robot_data *robots;
int nobstacles;
obst_data *obstacles;
planner_obj **probot;
planner_obj **pobstacle;
path_type cur_path_type=HOLO;
int curstep,maxstep;
int dx,dy,dth;
FILE *logfd=stderr;

/****************************************************************************
  Static Global Variables
****************************************************************************/

static MrmHierarchy mrm_id;
static char *mrm_vec[]={"xmplan.uid"};
static MrmCode mrm_class;
static MRMRegisterArg mrm_names[] = {
  {"File", (caddr_t)File },
  {"Load_Obstacles", (caddr_t)Load_Obstacles },
  {"Save_Obstacles", (caddr_t)Save_Obstacles },
  {"Load_Robots", (caddr_t)Load_Robots },
  {"Save_Robots", (caddr_t)Save_Robots },
  {"Load_Paths", (caddr_t)Load_Paths },
  {"Save_Paths", (caddr_t)Save_Paths },
  {"Refresh", (caddr_t)Refresh },
  {"Reset", (caddr_t)Reset },
  {"UnmanageCb", (caddr_t)UnmanageCb },
  {"Quit", (caddr_t)Quit },
  {"Verbosity_Set", (caddr_t)Verbosity_Set },
  {"Debug", (caddr_t)Debug },
  {"Really_Quit", (caddr_t)Really_Quit },
  {"Plan_Holo", (caddr_t)Plan_Holo },
  {"Plan_NHolo", (caddr_t)Plan_NHolo },
  {"Show_Holo_Path", (caddr_t)Show_Holo_Path },
  {"Show_NHolo_Path", (caddr_t)Show_NHolo_Path },
  {"Optimize_NHolo_Path", (caddr_t)Optimize_NHolo_Path },
  {"Animate_Holo_Path", (caddr_t)Animate_Holo_Path },
  {"Animate_NHolo_Path", (caddr_t)Animate_NHolo_Path },
  {"Change_Animate_Speed", (caddr_t)Change_Animate_Speed }
};

/***************************************************************************
  w_printf: print a string to a label widget
****************************************************************************/

void 
w_printf(Widget w,char *str) {

  XmString xmstr = XmStringCreateSimple(str);
  
  XtVaSetValues(w,XmNlabelString, xmstr,NULL);
  XmStringFree(xmstr);
}

/***************************************************************************
  Find_Name_Widget: Find the widget id for a given name string
****************************************************************************/

void 
Find_Name_Widget(Widget *w, char *wname) {

  
  if (((*w) = XtNameToWidget(shell,wname))==NULL) {
    fprintf(logfd,"Warning: Can't fine widget name %s\n",wname);
  }

}

/***************************************************************************
  UnmanageCb: unmanage the given widget
****************************************************************************/

void 
UnmanageCb(Widget w, XtPointer client_data, XtPointer call_data) {

  XtUnmanageChild(w);
}

/***************************************************************************
  Load_Robot: callback routine for loading robot data from a file
****************************************************************************/

void 
Load_Robots(Widget w, XtPointer client_data,
	    XmFileSelectionBoxCallbackStruct *cbs) {  

  char *filename;

  if (! XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename)) {
    perror("internal error");
    return; /* internal error */
  }

  Init_Robots(filename);
  Update_Canvas_Robots_Conf(robots);
  Update_Planner_Robots_Conf(robots);
  Display_Environment();
  XtFree(filename);

}

/***************************************************************************
  Load_Obstacle: callback routine for loading obstacle data from a file
****************************************************************************/

void 
Load_Obstacles(Widget w, XtPointer client_data,  
	      XmFileSelectionBoxCallbackStruct *cbs) {  

  char *filename;

  if (! XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename)) {
    perror("internal error");
    return; /* internal error */
  }

  Init_Obstacles(filename);
  Update_Canvas_Obst_Conf(obstacles);
  Update_Planner_Obst_Conf(obstacles);
  Display_Environment();
  XtFree(filename);

}

/***************************************************************************
  Save_Robot: callback routine for save robot data to a file
****************************************************************************/

void 
Save_Robots(Widget w, XtPointer client_data,  
	    XmFileSelectionBoxCallbackStruct *cbs) {  

  char *filename;

  if (! XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename)) {
    perror("internal error");
    return; /* internal error */
  }

  Write_Robots(filename,nrobots,robots);
  XtFree(filename);

}

/***************************************************************************
  Save_Obstacles: callback routine for saving obstacle data to a file
****************************************************************************/

void 
Save_Obstacles(Widget w, XtPointer client_data,  
	       XmFileSelectionBoxCallbackStruct *cbs) {  

  char *filename;

  if (! XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename)) {
    perror("internal error");
    return; /* internal error */
  }

  Write_Obstacles(filename,nobstacles,obstacles);
  XtFree(filename);

}

/***************************************************************************
  Load_Paths: callback routine for loading paths from a file
****************************************************************************/

void 
Load_Paths(Widget w, XtPointer client_data,  
	   XmFileSelectionBoxCallbackStruct *cbs) {  

  char *filename;

  if (! XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename)) {
    perror("internal error");
    return; /* internal error */
  }

  Read_Paths(filename,nrobots,robots);
  XtFree(filename);

}

/***************************************************************************
  Save_Paths: callback routine for saving paths to a file 
****************************************************************************/

void 
Save_Paths(Widget w, XtPointer client_data,  
	   XmFileSelectionBoxCallbackStruct *cbs) {  

  char *filename;

  if (! XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename)) {
    perror("internal error");
    return; /* internal error */
  }

  Write_Paths(filename,nrobots,robots);
  XtFree(filename);

}

/***************************************************************************
  Load: main callback routine for the load menu
****************************************************************************/

void 
File(Widget w, int *file_type, XtPointer call_data) {  

  switch(*file_type) {
  case k_load_robot:
    XtManageChild(loadRobotFileDialog);
    break;
  case k_load_obst:
    XtManageChild(loadObstFileDialog);
    break;
  case k_load_path:
    XtManageChild(loadPathFileDialog);
    break;
  case k_save_robot:
    XtManageChild(saveRobotFileDialog);
    break;
  case k_save_obst:
    XtManageChild(saveObstFileDialog);
    break;
  case k_save_path:
    XtManageChild(savePathFileDialog);
    break;
  default:
    fprintf(logfd,"Unknow file type");
    break;
  }
}

/***************************************************************************
  Init_Timeout_Procedure: initialize some timeout procedure
****************************************************************************/

void 
Init_Timeout_Procedure(){

  /* for picking and dragging an object */
  Init_Drag_Timeout();
  /* for animation */
  t_data.proc=(XtTimerCallbackProc)Animate;
  t_data.id=0;
  t_data.delay=150;

}

/***************************************************************************
  Show_Holo: callback routine for showing the trace of the robot along a path
****************************************************************************/

void 
Show_Holo_Path(Widget w, int *file_type, XtPointer call_data) {

  int i,max_step=0;

  if (!holo_path_exist) {
    w_printf(commandL,"No Holonomic Path exists\n");
    return;
  }
  
  for(i=0;i<nrobots;i++) {
    max_step=MAX(max_step,robots[i].pobj.holo_path.len);
  }
  for(i=0;i<max_step;i++) {
    Display_Robots_Path_Step(i,HOLO);
  }

}  

/***************************************************************************
  Show_Holo: callback routine for showing the trace of the robot along a path
****************************************************************************/

void 
Show_NHolo_Path(Widget w, int *file_type, XtPointer call_data) {

  int i,max_step=0;

  if (!nholo_path_exist) {
    w_printf(commandL,"No Holonomic Path exists\n");
    return;
  }
  
  for(i=0;i<nrobots;i++) {
    max_step=MAX(max_step,robots[i].pobj.nholo_path.len);
  }
  for(i=0;i<max_step;i++) {
    Display_Robots_Path_Step(i,NONHOLO);
  }

}  

/***************************************************************************
  Animate_Holo_Path: callback routine for animating the robot 
  for the current path
****************************************************************************/

void 
Animate_Holo_Path(Widget w, XtPointer dummy, XtPointer call_data) {

  int i;

  if (!holo_path_exist) {
    w_printf(commandL,"No Holonomic path exists\n");
    return;
  }

  cur_path_type=HOLO;
  curstep = 0;
  maxstep =0;
  for(i=0;i<nrobots;i++) {
    maxstep=MAX(maxstep,robots[i].pobj.holo_path.len);
  }
  
  Start_Timeout(&t_data);

}  

/***************************************************************************
  Animate_Holo_Path: callback routine for animating the robot 
  for the current path
****************************************************************************/

void 
Animate_NHolo_Path(Widget w, XtPointer dummy, XtPointer call_data) {

  int i;
  if (!nholo_path_exist) {
    w_printf(commandL,"No Nonholonomic path exists\n");
    return;
  }

  cur_path_type=NONHOLO;
  curstep = 0;
  maxstep =0;
  for(i=0;i<nrobots;i++) {
    maxstep=MAX(maxstep,robots[i].pobj.nholo_path.len);
  }
  
  Start_Timeout(&t_data);

}  

/***************************************************************************
  Animate: main timeout procedure for doing animation
****************************************************************************/

void 
Animate(timeout_data *to) {

  Stop_Timeout(to);
  if (curstep>=maxstep) {
    Display_Environment();
    return;
  }
  Display_Environment_WO_Flush();
  Display_Robots_Path_Step_On_Buffer(curstep,HOLO);
  /*Display_Path_Step(curpath->path[curstep],curcolor);*/
  XCopyArea(canvas_data.display,canvas_data.pmap,canvas_data.window,
            canvas_data.gc[bg],0,0,canvas_data.cdim.width,
	    canvas_data.cdim.height,0,0);
  XFlush(canvas_data.display);
  curstep++;
  Start_Timeout(to);

}  

/***************************************************************************
  Reset: redraw the screen with the current setting
****************************************************************************/

void 
Reset(Widget w,XtPointer client_data, XtPointer call_data) {

  Display_Environment();
}  

/***************************************************************************
  Refresh: refreshing the screen from the buffered pixmap
****************************************************************************/

void 
Refresh(Widget w, XtPointer client_data, XtPointer call_data) {

  Refresh_Environment();

}  

/***************************************************************************
  Quit: callback routine for exiting the program
****************************************************************************/

void 
Quit(Widget w, XtPointer client_data, XtPointer call_data) {

  if (confirm) {
    XtManageChild(confirmDialog); 
  } else {
    Really_Quit(0,NULL,NULL);
  }

}  

/***************************************************************************
  Verbosity_Set: callback routine for setting the printout verbosity
****************************************************************************/

void 
Verbosity_Set(Widget w, int *v, XtPointer call_data) {  

  fprintf(logfd,"Verbosity has been changed to %d\n",*v);
  verbosity=(*v);
}

/***************************************************************************
  Debug: callback routine for the debug menu
****************************************************************************/

void 
Debug(Widget w, int *v, XtPointer call_data) {  

  switch(*v) {
  case k_smooth:
    debug_smooth = !debug_smooth;
    V1(fprintf(logfd,"Debugging smoothing flag is turned ");
    fprintf(logfd,debug_smooth? "on\n":"off\n"));  
    break;
  case k_debug2:
    /* more debugging option */
    break;
  case k_debug3:
    /* more debugging option */
    break;
  default:
    break;
  }

}

/***************************************************************************
  Change_Animate_Speed: callback routine for changing the animation speed
****************************************************************************/

void 
Change_Animate_Speed(Widget w, XtPointer client_data,XtPointer call_data){

  int temp;

  XmScaleGetValue(w,&temp);
  t_data.delay=200.0*((100.0-temp)/100.0)+50;

}

/***************************************************************************
  Really_Quit: Routine to really exit the program
****************************************************************************/

void 
Really_Quit(Widget w, XtPointer client_data, XtPointer call_data) {  

  V1(fprintf(logfd,"+=============================================================+\n");
  fprintf(logfd,"| Thank you for using the motion planner developed \n");
  fprintf(logfd,"| in CS326 at the Stanford University.\n");
  fprintf(logfd,"+=============================================================+\n"));
  exit(0);
}

/***************************************************************************
  Plan_Holo: sample callback routine for doing holonomic planning
****************************************************************************/

void 
Plan_Holo(Widget w, XtPointer client_data, XtPointer call_data) {  

  int i;
  int old_len;
  char str[MAX_STR];

  holo_path_exist=FALSE;
  w_printf(commandL,"Search for a holonomic path ...\n");
  for(i=0;i<nrobots;i++) {
    V1(fprintf(logfd,"Planning for robot #%d\n",i));
    Holo_Plan(dx,dy,&robots[i].pobj,nobstacles,pobstacle,
	      &robots[i].pobj.holo_path);

    if (!debug_smooth) {
      do {
	old_len=robots[i].pobj.holo_path.len;
	V2(fprintf(logfd,"Old length=%d\n",old_len));
	Smooth1(&robots[i].pobj.holo_path);
	V2(fprintf(logfd,"New length=%d\n",robots[i].pobj.holo_path.len));
      } while(old_len!=robots[i].pobj.holo_path.len);
    }
    if (robots[i].pobj.holo_path.len != -1) {
      holo_path_exist=TRUE;
      sprintf(str,"A holonomic Path is Found for robot #%d\n",i);
    } else {
      sprintf(str,"No holonomic Path is Found for robot #%d\n",i);
    }
    w_printf(commandL,str);
  }

}
/***************************************************************************
  Plan_NHolo: sample callback routine for doing nonholonomic planning
****************************************************************************/

void 
Plan_NHolo(Widget w, XtPointer client_data, XtPointer call_data) {  

  int i;
  char str[MAX_STR];

  nholo_path_exist=FALSE;
  w_printf(commandL,"Search for a holonomic path ...\n");
  for(i=0;i<nrobots;i++) {
    V1(fprintf(logfd,"Planning for robot #%d\n",i));
    NHolo_Plan(dx,dy,&robots[i].pobj,nobstacles,pobstacle,
	       &robots[i].pobj.nholo_path);
    if (robots[i].pobj.nholo_path.len != -1) {
      nholo_path_exist=TRUE;
      sprintf(str,"A nonholonomic Path is Found for robot #%d\n",i);
    } else {
      sprintf(str,"A nonholonomic Path is Found for robot #%d\n",i);
    }
    w_printf(commandL,str);

  }

}


/***************************************************************************
  Optimize_NHolo_Path: sample callback routine for optimizing 
  nonholonomic paths
****************************************************************************/

void 
Optimize_NHolo_Path(Widget w, XtPointer client_data, XtPointer call_data) {  

  register int i;
  w_printf(commandL,"Optimizing Path ...\n");
  for (i=0;i<nrobots;i++) {
    Optimize(dx,dy,&robots[i].pobj,nobstacles,pobstacle,
			&robots[i].pobj.nholo_path);
  }
  w_printf(commandL,"Optimization is done.\n");

}

/*****************************************************************************
  Init_Robot: Initialize the data struct for the robot from a file
*****************************************************************************/

void 
Init_Robots(char *robot_filename) {

  int i;

  Read_Robots(robot_filename,&nrobots,&robots);

  for(i=0;i<nrobots;i++) {
    robots[i].id=0;
    robots[i].cobj.type=ROBOT_INIT;
    robots[i].cgobj.type=ROBOT_GOAL;
    robots[i].cobj.orig_poly_ptr = &robots[i].pobj.orig_poly;
    robots[i].cgobj.orig_poly_ptr = &robots[i].pobj.orig_poly;
    /* Initialize the color for your roobts here 
       if you want them to be different */
    robots[i].cobj.fill_c=green;
    robots[i].cobj.outline_c=black;
    robots[i].cgobj.fill_c=red;
    robots[i].cgobj.outline_c=black;
    robots[i].animate_color=blue;
    robots[i].trace_color=orange;
  }

}

/*****************************************************************************
  Init_Obstacles:Initialize the data struct for the obstacles from a file
*****************************************************************************/

void 
Init_Obstacles(char *obst_filename) {

  int i;

  Read_Obstacles(obst_filename,&nobstacles,&obstacles);

  for(i=0;i<nobstacles;i++) {
    obstacles[i].id=i;
    obstacles[i].cobj.type=OBST;
    obstacles[i].cobj.outline_c=black;
    obstacles[i].cobj.fill_c=black;
    obstacles[i].cobj.orig_poly_ptr = &obstacles[i].pobj.orig_poly;
  }
  
}

/*****************************************************************************
  Init_Paths:Initialize the path data struct
*****************************************************************************/

void 
Init_Paths(char *path_filename) {

  if ((nrobots!=0)&&(path_filename[0]!='\0')) {
    Read_Paths(path_filename,nrobots,robots);
  }
  
}

/****************************************************************************
  Init_Widget: Find widget id's for the given names of all intereted widgets
****************************************************************************/

void 
Init_Widget(void) {

  Find_Name_Widget(&commandL,"*commandLabel");
  Find_Name_Widget(&canvas,"*canvas");
  Find_Name_Widget(&planL,"*planLabel");
  Find_Name_Widget(&infoDialog,"*infoDialog");
  Find_Name_Widget(&confirmDialog,"*confirmDialog");
  Find_Name_Widget(&loadRobotFileDialog,"*loadRobotFileDialog");
  Find_Name_Widget(&loadObstFileDialog,"*loadObstacleFileDialog");
  Find_Name_Widget(&loadPathFileDialog,"*loadPathFileDialog");
  Find_Name_Widget(&saveRobotFileDialog,"*saveRobotFileDialog");
  Find_Name_Widget(&saveObstFileDialog,"*saveObstacleFileDialog");
  Find_Name_Widget(&savePathFileDialog,"*savePathFileDialog");

}

/****************************************************************************
  main: main function call
****************************************************************************/

int 
main(int argc,char **argv) {

  Display *display;
  Widget main_window;
  arg *args;
  
  /* parsing the command line arguments */
  args=parse_args(argc,argv);

  if (!GetIntOpt(args,'v',&verbosity,1,0,4)) {
    V1(fprintf(stderr,"Use default verbosity %d\n",verbosity));
  } else {
    V1(fprintf(stderr,"Verbosity:%d\n",verbosity));
  }

  if (!GetStrOpt(args,'o',obst_filename,"data/obstacle.dat",1,256)) {
    V1(fprintf(logfd,"Use default obstacle description file name %s\n",
               obst_filename));
  } else {
    V1(fprintf(logfd,"Use specified obstacle description file name %s\n",
               obst_filename));
  }
  if (!GetStrOpt(args,'r',robot_filename,"data/robot.dat",1,256)) {
    V1(fprintf(logfd,"Use default robot description file name %s\n",
               robot_filename)); 
  } else {
    V1(fprintf(logfd,"Use specified robot description file name %s\n",
               robot_filename)); 
  }
  if (args[chindex('p')].ch) {
    if (args[chindex('p')].str==NULL) {
      strcpy(path_filename,"data/path.dat");
      V1(fprintf(logfd,"Use default path file: %s\n",
		 path_filename)); 
    } else {
      if (!GetStrOpt(args,'p',path_filename,"data/path.dat",1,256)) {
	V1(fprintf(logfd,"Use default path file: %s\n",
		   path_filename)); 
      } else {
	V1(fprintf(logfd,"Use specified path file: %s\n",
		   path_filename)); 
      }
    }
  }

  if (!GetIntOpt(args,'x',&dx,DX,0,1000)) {
    V1(fprintf(logfd,"Use default x dimension %d\n",dx));
  } else {
    V1(fprintf(logfd,"X dimension:%d\n",dx));
  }

  if (!GetIntOpt(args,'y',&dy,DY,0,1000)) {
    V1(fprintf(logfd,"Use default y dimension %d\n",dy));
  } else {
    V1(fprintf(logfd,"Y dimension:%d\n",dy));
  }

  if (!GetIntOpt(args,'t',&dth,DTH,0,1000)) {
    V1(fprintf(logfd,"Use default theta dimension %d\n",dth));
  } else {
    V1(fprintf(logfd,"Theta dimension:%d\n",dth));
  }

  if (!GetFloatOpt(args,'s',&wscale,1.0,0.0,1000.0)) {
    V1(fprintf(logfd,"Use default world scale: %f\n",wscale));
  } else {
    V1(fprintf(logfd,"World Scale:%f\n",wscale));
  }

  Init_Robots(robot_filename);
  Init_Obstacles(obst_filename);

  XtToolkitInitialize();
  MrmInitialize ();
  app_context = XtCreateApplicationContext();
  display = XtOpenDisplay(app_context, NULL, argv[0], "Xmplan",
                          NULL, 0, &argc, argv);
  if (display == NULL) {
    fprintf(logfd, "%s:  Can't open display\n", argv[0]);
    exit(1);
  }

  shell = XtAppCreateShell(argv[0], NULL, applicationShellWidgetClass,
                           display, NULL, 0);
  if (MrmOpenHierarchy (1, mrm_vec, NULL, &mrm_id) != MrmSUCCESS) exit(0);

  MrmRegisterNames(mrm_names, XtNumber(mrm_names));

  if (MrmFetchWidget(mrm_id, "mainWindow", shell, &main_window, &mrm_class)
      != MrmSUCCESS)  perror("mainWindow not found");

  
  /* Initialze the interface widget stuff */
  Init_Widget();
  XtManageChild(main_window);
  XtRealizeWidget(shell);

  /* Init_Planner must be called before Init_Graphics */
  Init_Planner(dx,dy,dth);

  /* Initialize the graphics on canvas*/
  Init_Graphics();
  Init_Paths(path_filename);

  Display_Environment();

  XtAppMainLoop(app_context);
  
  return(TRUE);
}

