/* Copyright (c) 1995 by Stanford University.  All Rights Reserved */

/***************************************************************************
   NAME
     mif_plan.c
   PURPOSE
     Sample/empty planning routines for CS326
   NOTES
     
   AUTHOR
     Tsai-Yen Li (li@flamingo.Stanford.EDU)
   HISTORY
     Tsai-Yen Li - Apr 10, 1995: Created.
***************************************************************************/
#ifdef RCS
#ifndef lint
 static char rcsid[]="$Header: /u/li/Proj/mif/RCS/mif_plan.c,v 1.3 1995/04/15 22:07:52 li Exp li $";
#endif 
#endif 
/* $Log: mif_plan.c,v $
 * Revision 1.3  1995/04/15  22:07:52  li
 * add path interface
 *
 * Revision 1.2  1995/04/14  01:33:14  li
 * Deal with multiple robots
 *
 * Revision 1.1  1995/04/12  08:00:18  li
 * Initial revision
 *
 * Revision 1.1  1995/04/11  21:15:40  li
 * Initial revision
 * */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/param.h>
#include <X11/keysym.h>
#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Mrm/MrmPublic.h>
#include "planner_header.h"
#include "mif_header.h"
#include "mif_proto.h"

/***************************************************************************
  Checkout_Path_Memory
****************************************************************************/

void
Check_Path_Memory(objpath *cpath,int len) {

  if (len<=0) return;
  if ((len>cpath->len)||(cpath->path==NULL)) {
    if ((cpath->len==0)&&(cpath->path!=NULL))
      cpath->path=NULL;
    if (cpath->path!=NULL) {
      free(cpath->path);
    }
    cpath->path=(Fq *)calloc(len,sizeof(Fq));
    if (cpath->path==NULL) {
      NO_MORE_MEMORY("Check_Path_Memory");
    }
    cpath->size=len;
  }
}

/***************************************************************************
  Interpolate_Conf
****************************************************************************/

void 
Interpolate_Conf(Fq A, Fq B,int interval,Fq *st) {

  register int i;
  
  for(i=0;i<interval;i++) {
    st[i].x= A.x + ((B.x - A.x)*(float)i)/(float)interval;
    st[i].y= A.y + ((B.y - A.y)*(float)i)/(float)interval;
    if((B.th - A.th) > dth*0.5)
      st[i].th = A.th - ((dth - (B.th - A.th))*(float)i)/(float)interval;
    else if((B.th - A.th) < -dth*0.5)
      st[i].th = A.th +((dth + (B.th - A.th))*(float)i)/(float)interval ;
    else
      st[i].th = A.th +((B.th-A.th)*(float)i)/(float)interval ;
    
    NORMALIZE(st[i].th,dth);
  }

}

/****************************************************************************
  Print_Fpolygons
****************************************************************************/

void
Print_Fpolygons(int npoly,Fpolygon *poly) {

  int i,j;

  for (i=0;i<npoly;i++) {
    fprintf(logfd,"Polygon #%d:\n",i);
    for (j=0;j<=poly[i].nvert;j++) {
      fprintf(logfd,"%d: %f %f\n",j,poly[i].vert[j].x,poly[i].vert[j].y);
    }
  }
}

/****************************************************************************
  Print_Robot_Polygons
****************************************************************************/

void
Print_All_Polygons(planner_obj *robot,int nobst,planner_obj **obsts) {

  int i;

  fprintf(logfd,"Robot polygons:\n");
  Print_Fpolygons(robot->npoly,robot->poly);
  fprintf(logfd,"Obstacle polygons:\n");
  for(i=0;i<nobst;i++) { 
    fprintf(logfd,"Obstacle #%d:\n",i);
    Print_Fpolygons(obsts[i]->npoly,obsts[i]->poly);
  }

}

/****************************************************************************
  Holo_Plan
****************************************************************************/

int 
Holo_Plan(int dx,int dy,planner_obj *robot,int nobst,planner_obj **obstacle,
	  objpath *holo_path) {


  V2(Print_All_Polygons(robot,nobst,obstacle));
  holo_path->len=11;
  Check_Path_Memory(holo_path,holo_path->len);
  Interpolate_Conf(robot->init,robot->goal,10,holo_path->path);
  holo_path->path[holo_path->len-1]=robot->goal;
  return(TRUE);

}

/****************************************************************************
  NHolo_Plan
****************************************************************************/

int 
NHolo_Plan(int dx,int dy,planner_obj *robot,int nobst,planner_obj **obstacle,
	   objpath *nholo_path) {


  nholo_path->len=0;

  return(TRUE);

}

/****************************************************************************
  Optimize
****************************************************************************/

int 
Optimize(int dx,int dy,planner_obj *robot,int nobst,
		    planner_obj **obstacle,objpath *nholo_path) {

  return(TRUE);

}

/****************************************************************************
  Smooth1
****************************************************************************/

void 
Smooth1(objpath *pX) {

  

}


