/* Copyright (c) 1995 by Stanford University.  All Rights Reserved */

/***************************************************************************
   NAME
     pheader.h
   PURPOSE
     
   NOTES
     
   AUTHOR
     Tsai-Yen Li (li@flamingo.Stanford.EDU)
   HISTORY
     Tsai-Yen Li - Apr 10, 1995: Created.
***************************************************************************/
#ifdef RCS
#ifndef lint
 static char rcsid[]="$Header: /u/li/Proj/mif/RCS/planner_header.h,v 1.3 1995/04/14 01:33:14 li Exp $";
#endif 
#endif 
/* $Log: planner_header.h,v $
 * Revision 1.3  1995/04/14  01:33:14  li
 * Deal with multiple robots
 *
 * Revision 1.3  1995/04/14  01:33:14  li
 * Deal with multiple robots
 *
 * Revision 1.2  1995/04/12  08:00:18  li
 * add comments and fix some bugs
 *
 * Revision 1.1  1995/04/11  21:15:40  li
 * Initial revision
 * */

/* defining constants */

#define false		0
#define true		1
#ifndef FALSE
#define FALSE		0
#endif
#ifndef TRUE
#define TRUE		1
#endif
#define DX 128
#define DY 128
#define DTH 96
#define DTH2 48

#define MAXOBST	        64
#define GOAL		0
#define OBSTACLE	255
#define NOTSET		255
#define UNVISITED       254
#define NOTEST          254
#define MAX_VERTEX 20
#define MAX_POT 100000
#define MAX_NODE 100000

#ifndef MIN
#define MIN(x,y) ((x > y)? y : x)
#endif
#ifndef MAX
#define MAX(x,y) ((x > y)? x : y)
#endif

#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif
#ifndef TWOPI
#define TWOPI 6.28218530717958647692
#endif

/* useful macros */

/* for versosity level */
#define V1(p) if(verbosity > 0) {p;}
#define V2(p) if(verbosity > 1) {p;}
#define V3(p) if(verbosity > 2) {p;}
#define V4(p) if(verbosity > 3) {p;}

#define RNDINT(i) (int)(i+0.5)
#define NORMALIZE(a,ma) if(a < 0) a += ma; if (a >= ma) a -= ma;
#define TRUNC(a,b,c) (((a) < b) ? b : ((a) >= c) ? (c-1) : (a))
#define TRUNC0(x,y) (((x) < 0) ? 0 : ((x) >= y) ? (y-1) : (x))
#define IN_BOUND(x,y,dx,dy) ((x>0)&&(x<dx-1)&&(y>0)&&(y<dy-1))
#define TIME(p) {clock_t start_t,end_t;\
                   start_t=clock(); p; end_t=clock();\
                     printf("%f sec.\n",(end_t-start_t)/1000000.0);\
                     }\


/* defining data structure */

/* integer 3d configuration */
typedef struct _Iq{
  int x,y,th;
} Iq;

/* floating-point 3d configuration */
typedef struct _Fq{
  float x,y,th;
} Fq;

/* integer 2d point */
typedef struct _ipoint {
  int x,y;
} Ipoint;

/* floating-point 2d point */
typedef struct _fpoint {
  float x,y;
} Fpoint;

/* integer polygon */
typedef struct _ipolygon {
  int nvert;
  Ipoint *vert;
} Ipolygon;

/* floating-point polygon */
typedef struct _fpolygon {
  int nvert;
  Fpoint *vert;
} Fpolygon;

/* 2d bytemap */
typedef struct _Bytemap2d {
  short dim1,dim2;
  unsigned char **bmap;
} Bytemap2d;

/* 2d bitmap */
typedef struct _Bitmap2d {
  short dim1,dim2;
  int **bmap;
} Bitmap2d;

/* 3d bitmap */
typedef struct _Bitmap3d {
  short dim1;
  short dim2;
  short dim3;
  int ***bmap;
} Bitmap3d;

/* 3d bytemap */
typedef struct _Bytemap3d {
  short dim1;
  short dim2;
  short dim3;
  unsigned char ***bmap;
} Bytmap3d;

/* path for the robot */
typedef struct _objpath{
  int size;   /* allocated size for the array below */
  int len; /* length of the current path */
  Fq *path;   /* array of configurations */
} objpath;

/* main data struct for an object (robot/obstacle) */
typedef struct _planner_obj {
  int npoly; /* number of polygons */
  Fpolygon *orig_poly; /* polygons for (0,0,0) configuration */
  Fpolygon *poly;      /* polygons for the current configuration */
  Fq init; /* initial configuration */
  Fq goal; /* goal configuation */
  objpath holo_path;
  objpath nholo_path;
} planner_obj;

/* external global variables */

extern int dx,dy,dth;
extern int nrobots,nobstacles;
extern planner_obj **pobstacle,**probot;
extern float pc_scale,cp_scale,dth2rad,rad2dth,delta_th;
extern int verbosity;

/* external system function proto types */

extern clock_t times();
extern unsigned sleep();
extern void bcopy();
extern int fprintf();
extern void perror();
extern int fgetc();
extern int fscanf();
extern int fclose();
/*extern void *memset();*/
